/**
 * @(#)tsidexaminer.cpp 2000/11/01
 *
 * @author Ice
 * @version 1.01
 *
 * ICE Team free software group
 *
 * This file is part of TSID.
 * See README for copyright notice.
 *
 * note: some comments are in the style of Javadoc of Java language!
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 *  02111-1307  USA.
 */

#ifndef _TSIDEXAMINER
#define _TSIDEXAMINER

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <pwd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <dirent.h>

#include "tsidexaminer.h"

/**
 * Construnct the TSIDExaminer
 */
TSIDExaminer::TSIDExaminer() {
  int handle;

  SDEBUG("Construct the TSIDExaminer\n");

  versionString="tsidexaminer 1.01 by Ice Team Free Software Group (C) 2000-2001\n";

  // base directory is not directly visible
  TDir=".tsid_hvsc";

  homeDir=NULL;
  homeTDir=NULL;
  startList=NULL;
  endList=NULL;
  startList2=NULL;
  endList2=NULL;

  // read the home directory of user
  if (!getHomeDir()) {
    cerr << "Error: not home directory found\n";
  }

  homeTDir=new char[strlen(homeDir)+strlen(TDir)+2];
  if (homeTDir==NULL) {
      cerr << "Error allocating memory\n";
  }
  sprintf(homeTDir, "%s/%s",homeDir, TDir);


  // verify if base directory tree is present
  if ((handle=open(homeTDir, O_RDONLY))==-1) {
                                              // not present
    cerr << "Warning: TSID directory not present\n";
  } else {                                    // present, close handle
      close(handle);
    }

  SDEBUG("home dir= " << homeDir << endl);
  SDEBUG("home TSID dir= " << homeTDir << endl);
}

/**
 * Destruct the TSID
 * Note: if you have called a <code>analiseTSid</code>, then you must call
 * <code>mergeSongList</code> and <code>mergeAuthorList</code> in this
 * sequence before calling this destructor for a correct termination.
 */
TSIDExaminer::~TSIDExaminer() {
  SDEBUG("Destruct the TSIDExaminer\n");

  // free the memory used by the list
  if (startList != NULL) {
    tuneItem *pTmp;                          // temp pointer

    while (startList->next != NULL) {
      pTmp=startList;                        // remember the start
      // this instruction is unsafe if the
      // expected flow is not followed
      delete pTmp->fileName;

      startList=startList->next;             // the new list
      delete pTmp;                           // release memory
    }

    startList=NULL;
    endList=NULL;
  }

  // free the memory used by the path list
  if (startList2 != NULL) {
    pathItem *pTmp;                          // temp pointer

    while (startList2->next != NULL) {
      pTmp=startList2;                       // remember the start
      startList2=startList2->next;           // the new list
      delete pTmp;                           // release memory
    }

    startList2=NULL;
    endList2=NULL;
  }

  delete(homeDir);
  delete(homeTDir);
}

/**
 * Get home directory of user
 * The private variables homeDir is chaged
 * Note: this code is taken and modified from XMame!!!
 *
 * @return true if operation is ok
 */
bool TSIDExaminer::getHomeDir() {
  struct passwd *pw;

  lastError=NO_TSIDE_ERROR;  // no error

  pw=getpwuid(getuid());
  if(!pw) {
      cerr << "Who are you? Not found in passwd database!!\n";
      lastError=NO_USER;  // no user found
      return false;
  }

  homeDir=new char[strlen(pw->pw_dir)+1];
  if (homeDir==NULL) {
      cerr << "Error allocating memory\n";
      lastError=NO_MEMORY;
      return false;
  }
  strcpy(homeDir, pw->pw_dir);

  return true;
}

/**
 * Analise the TSID directory for sidt files.
 */
void TSIDExaminer::analiseTSid() {
  // recurse into TSID directory
  recurse(homeTDir);
}

/**
 * Sort list in decrescent order using time or number field
 *
 * @param type the field to use
 */
void TSIDExaminer::sortList(int type) {
  tuneItem *actual;        // actual pointer
  tuneItem *start;         // actual start pointer (upper, all are ordered)
  tuneItem tmp;            // tmp for swap

  SDEBUG("Sort List\n");

  // swap element macro
  #define swap(a,b) \
     tmp.pathName=a->pathName; \
     tmp.fileName=a->fileName; \
     tmp.song=a->song; \
     tmp.second=a->second; \
     tmp.number=a->number; \
     a->pathName=b->pathName; \
     a->fileName=b->fileName; \
     a->song=b->song; \
     a->second=b->second; \
     a->number=b->number; \
     b->pathName=tmp.pathName; \
     b->fileName=tmp.fileName; \
     b->song=tmp.song; \
     b->second=tmp.second; \
     b->number=tmp.number;

  // use a bubble sort algorithm (oh, what I see)
  if (startList==NULL) return;               // list has no element
  if (startList->next==NULL) return;         // list has one element

  for (start=startList; start->next!=NULL; start=start->next)
    for (actual=start->next; actual!=NULL; actual=actual->next) {
      switch (type) {
        case SORT_FOR_TIME:
          if (actual->second > start->second) {
             swap(actual, start);
          }
          break;
        case SORT_FOR_NUMBER:
          if (actual->number > start->number) {
             swap(actual, start);
          }
          break;
        // default: do nothing
      }
    }
}

/**
 * Merge the equal song in the list
 * The tunes of the same song/author are merge in one
 */
void TSIDExaminer::mergeSongList() {
  tuneItem *actual;        // actual pointer
  tuneItem *start;         // actual start pointer (upper, all are merged)
  tuneItem *prev;          // previous element of actual

  SDEBUG("Merge Song List\n");

  // use a bubble sort algorithm (oh, what I see)
  if (startList==NULL) return;               // list has no element
  if (startList->next==NULL) return;         // list has one element

  for (start=startList; start->next!=NULL; start=start->next) {
    for (actual=start->next, prev=start;
         actual!=NULL;
         prev=actual, actual=actual->next) {
      if (strequ(actual->pathName, start->pathName)) {
        if (strequ(actual->fileName, start->fileName)) {
          // merge the actual element to the start element
          start->second+=actual->second;
          start->number+=actual->number;
          start->song=0;

          // remove actual from list
          prev->next=actual->next;
          delete actual;
          actual=prev;
        }
      }
    }
    // the list is modified, so this is needed
    if (start->next==NULL) break;
  }
}

/**
 * Merge the songs of the same author in the list
 * The tunes of the same author are merge in one
 * Note: you must call <code>mergeSongList</code> before use this methods.
 */
void TSIDExaminer::mergeAuthorList() {
  tuneItem *actual;        // actual pointer
  tuneItem *start;         // actual start pointer (upper, all are merged)
  tuneItem *prev;          // previous element of actual
  char *tmpStr="Various";

  #define substName \
    delete start->fileName; \
    start->fileName=new char[strlen(tmpStr)+1]; \
    strcpy(start->fileName, tmpStr); \
    start->song=0;

  SDEBUG("Merge Author List\n");

  // use a bubble sort algorithm (oh, what I see)
  if (startList==NULL) return;               // list has no element
  if (startList->next==NULL) {               // list has one element
    substName;
    return;
  }

  for (start=startList; start->next!=NULL; start=start->next) {
    for (actual=start->next, prev=start;
         actual!=NULL;
         prev=actual, actual=actual->next) {
      if (strequ(actual->pathName, start->pathName)) {
        // merge the actual element to the start element
        start->second+=actual->second;
        start->number+=actual->number;
        // remove actual from list
        prev->next=actual->next;
        delete actual->fileName;
        delete actual;
        actual=prev;
      }
    }
    substName;

    // the list is modified, so this is needed
    if (start->next==NULL) break;
  }
  substName;
}

/**
 * Recurse through the given directory and manage the founded files.
 * This code is taken and modified from gzip sources
 *
 * @param dir the base directory to recurse in
 */
void TSIDExaminer::recurse(char *dir) {
  dirent *dp;
  struct stat st;
  DIR    *dirp;
  char   nbuf[MAX_PATH_LEN];
  int    len;

  SDEBUG("Recurse into directory " << dir << endl);

  lastError=NO_TSIDE_ERROR;                  // no error

  dirp=opendir(dir);

  if (dirp==NULL) {
    cerr << dir << " unreadable\n";
    lastError=DIR_READ;                      // read error of directory
    return;
  }

  while ((dp=readdir(dirp)) != NULL) {
    if (strequ(dp->d_name,".") || strequ(dp->d_name,"..")) {
      continue;
    }
    len=strlen(dir);
    if (len+NLENGTH(dp)+1 < MAX_PATH_LEN-1) {
      strcpy(nbuf,dir);
      if (len != 0                 // dir = "" means current dir on Amiga
#ifdef PATH_SEP2
        && dir[len-1] != PATH_SEP2
#endif
#ifdef PATH_SEP3
        && dir[len-1] != PATH_SEP3
#endif
        ) {
        nbuf[len++]=PATH_SEP;
      }
      strcpy(nbuf+len, dp->d_name);

      stat(nbuf, &st);

      if (S_ISDIR(st.st_mode)) {
        recurse(nbuf);                       // recurse into directory
      }

      if (S_ISREG(st.st_mode)) {
        addFileToList(nbuf);                 // add the file
      }
    } else {
        cerr << dir << "/" << dp->d_name << ": pathname too long\n";
      }
  }
  closedir(dirp);
}

/**
  * Insert the information of SIDT file into the list
  *
  * @param nfile name of file
  */
void TSIDExaminer::addFileToList(char *nfile) {
  SIDT *pSidt=new SIDT(nfile);  // pointer to sidt file manager
  tuneItem *pItem;              // a item for the list
  char *pPath;                  // pointer to path of file
  char *pName;                  // pointer to name of file

  SDEBUG("Add information from file " << nfile << endl);

  if (!pSidt->read()) {
    cerr << "Error reading SIDT information\n";
    return;
  }

  pPath=addPathToList2(getPath(nfile));      // get the path and insert in list
  pName=getName(nfile);                      // get the name of file

  pSidt->initForGet();                       // init for getting next subtune


  pItem=new tuneItem;

  // scan all the subtunes that are present
  while (pSidt->getNext(&pItem->song, &pItem->second, &pItem->number)) {

    SDEBUG("Add song #" << pItem->song << endl);

    if (startList==NULL) {                   // list empty ?
      startList=pItem;                       // this is the first element
      endList=pItem;                         // and the last too
    } else {                                 // list not empty
        endList->next=pItem;                 // next point to the last
        endList=pItem;                       // this is the last
      }
    // pItem must be completed with the path and file name
    pItem->pathName=pPath;
    pItem->fileName=pName;
    pItem->next=NULL;                        // there's not other
    pItem=new tuneItem;
  }

  delete pItem;
  delete pSidt;
}

/**
 * Add the path to the list of paths
 *
 * @param path the path to add (as a copy to use)
 * @return the pointer to the path in the list
 */
char *TSIDExaminer::addPathToList2(char *path) {
  pathItem *pItem;                           // a item for the list
  pathItem *actual;                          // actual element in the list

  SDEBUG("Add path: " << path << " to list of path\n");

  // scan all the list for finding if it is present
  if (startList2==NULL) {                    // list empty ?

      pItem=new pathItem;                    // create a new item
      pItem->pathName=path;
      startList2=pItem;                      // this is the first element
      endList2=pItem;                        // and the last too
      pItem->next=NULL;                      // there's not other
      return path;

    } else {                                 // list not empty

        actual=startList2;                   // start the list
        while (actual!=NULL) {
          // the list is not ordered, so simple implementation

          if (strcmp(actual->pathName, path)==0) {
            delete path;
            return actual->pathName;
          }
          actual=actual->next;
        }

        // add element at the end of list
        pItem=new pathItem;                  // create a new item
        pItem->pathName=path;
        pItem->next=NULL;                    // there's not other
        endList2->next=pItem;                // next point to the last
        endList2=pItem;                      // this is the last

        return path;
      }
}

/**
 * Show the list to the passed output stream
 * The stream must ready for use
 *
 * @param out the output stream pointer
 */
void TSIDExaminer::showList(ostream *out) {
  tuneItem *actual;              // pointer to the list
  char *tmpStr="Various";
  int step=strlen(homeTDir);     // this is not to show
  int position=1;                // position in list
  unsigned long totalSec=0;      // total listen seconds
  unsigned long totalNum=0;      // total number of listen

  SDEBUG("Show list to passed stream\n");

  actual=startList;
  while (actual!=NULL) {
    // calculate statistic
    totalSec+=actual->second;
    totalNum+=actual->number;

    // show the information to the output stream

    *out << position++ << "\t";
    *out << actual->second << "\t";
    *out << actual->number << "\t";

    *out << ((actual->pathName)+step) << " ";

    if (strcmp(actual->fileName, tmpStr)!=0) {
      *out << actual->fileName << " ";
    }

    if (actual->song!=0) {
      *out << "#" << actual->song;
    }

    *out << endl;

    actual=actual->next;                     // go to next element
  }
  *out << "*** STATISTICS ***\n";
  *out << "LISTEN SECONDS: " << totalSec << endl;
  *out << "NUMBER of LISTEN: " << totalNum << endl;
}

/**
 * Get path of file (as a new copy)
 *
 * @param nfile the name of the file
 * @return the pointer to a new copy of the file path
 */
char *TSIDExaminer::getPath(char *nfile) {
  char size;                                 // size of string

  SDEBUG("Get path from file name: ");

  char *pnt=&nfile[strlen(nfile)];           // pointer to the end of string

  while (*pnt != PATH_SEP) {                 // find the separator
    pnt--;                                   // decrement pointer
    if (pnt<=nfile) break;                   // avoid endless loop
  }

  char *out=new char[(size=(pnt-nfile))+1];  // create the empty string
  strncpy(out, nfile, size);                 // copy the string

  out[size]='\0';                            // terminate the string

  SDEBUG(out << endl);

  return out;
}

/**
 * Get name of the file (as a new copy)
 *
 * @param nfile the name of the file
 * @return the pointer to a new copy of the file name
 */
char *TSIDExaminer::getName(char *nfile) {
  char size;                    // size of string

  SDEBUG("Get name from path-file name: ");

  char *pnt=&nfile[size=strlen(nfile)];      // pointer to the end of string
  while (*pnt != PATH_SEP) {                 // find the separator
    pnt--;                                   // decrement pointer
    if (pnt<=nfile) break;                   // avoid endless loop
  }

  size-=(pnt-nfile);
  char *out=new char[size];                  // create the empty string
  strncpy(out, ++pnt, size);                 // copy the string

  SDEBUG(out << endl);

  return out;
}

#endif // _TSIDExaminer.CPP


